#include <fstream.h>
#include <cassert>
#include <cmath>
#include "Math.hh"
#include "Random.hh"
#include "Anneal.hh"

Anneal::Anneal(funcPtr objFunc, perturbPtr pertFunc,
	       ReductionSchedule reduce, AcceptanceCriteria accept,
	       int numTemps, int numIters) {
#ifdef DEBUG
  assert(numTemps > 1);
  assert(numIters > 1);
#endif
  this->objFunc  = objFunc;
  this->pertFunc = pertFunc;
  this->reduce   = reduce;
  this->accept   = accept;
  this->numTemps = numTemps;
  this->numIters = numIters;
}

double
Anneal::anneal(Vector& point) {
#ifdef DEBUG
  ofstream outStream("Anneal.log");
  outStream.precision(20);
#endif
  double tempMult = 0.0, startTemp = 1.0, stopTemp = 0.01;
  if(reduce == GEOMETRIC)
    tempMult = Math::safeExp(log(stopTemp / startTemp) / (numTemps - 1.0));
  else if(reduce == FAST)
    tempMult = (startTemp - stopTemp) / (stopTemp * (numTemps - 1.0));

  double fvalbest, fvalcurr, fval;
  Vector bestpoint(point), newpoint(point.getSize());
  fvalcurr = fvalbest = objFunc(point);

  int n = 0;
  double fsum = 0.0, fsqrsum = 0.0;
  double prob = 0.0, temp = 1.0, stdev = 1.0;
  Random ran(797L);

  for(int i = 0; i < numTemps; i++) {
    for(int j = 0; j < numIters; j++) {

      pertFunc(point, newpoint, temp);
      fval = objFunc(newpoint);

      n++;
      fsum += fval;
      fsqrsum += fval * fval;

      if(fval < fvalbest) {
	fvalbest  = fval;
	bestpoint = newpoint;
	j -= 1;
	if(j < 0) j = 0;
      }

      if(accept == METROPOLIS)
	prob = Math::safeExp((fvalcurr - fval) / (stdev * temp));
      else if(accept == SZU)
	prob = 1.0 / (1.0 + Math::safeExp((fval - fvalcurr) / (stdev * temp)));

      if(ran.uniform() < prob) {
	fvalcurr = fval;
	point    = newpoint;
      }
    }

    stdev = sqrt((fsqrsum - (fsum * fsum) / (double)n) / (double)n);
    n = 0; fsum = fsqrsum = 0.0;

    fvalcurr = fvalbest;
    point    = bestpoint;

    if(reduce == GEOMETRIC)
      temp *= tempMult; 
    else if(reduce == FAST)
      temp = startTemp / (1.0 + tempMult * (i + 1.0));
#ifdef DEBUG
    outStream << "Total iterations:    " << (i+1)*numIters << endl;
    outStream << "Current temperature: " << temp           << endl;
    outStream << "Best function value: " << fvalbest       << endl;
    outStream << "At: "                  << bestpoint      << endl;
    outStream << endl;
#endif
  }
#ifdef DEBUG
  outStream.close();
#endif
  point = bestpoint;
  return fvalbest;
}

Anneal::~Anneal() {
}

void 
perturb(const Vector& point, Vector& newpoint, double temp) {
  static Random ran(1283L);
  static double stdev = 100.0;
  for(int i = 0; i < point.getSize(); i++)
    newpoint[i] = point[i] + temp * stdev * ran.normal();
}
