#include "Math.hh"
#include "Covariance.hh"

////////// Covariance //////////

Covariance::Covariance(int dimensions) : SymmetricMatrix(dimensions) {
}

Covariance::~Covariance() {
}

////////// ARD //////////

inline
double 
ARD::ard(const Vector& p, const Vector& q, const Vector& param) const {
  return param[v] * Math::safeExp(-0.5 * p.squaredNorm(q, param));
}

ARD::ARD(const Matrix& data) : train(data), v(data.getCols()) {
}

ARD::~ARD() {
}

////////// ARD + Linear ////////// 

inline
double 
ARDLinear::linear(const Vector& p, const Vector& q, const Vector& param) const {
  return param[a0] + param[a1] * p.dotProduct(q);
}

ARDLinear::ARDLinear(const Matrix& data) : ARD(data), a0(v+1), a1(v+2) {
}

ARDLinear::~ARDLinear() {
}

////////// ARD + Linear + Noise //////////

inline
int 
ARDLinearNoise::getNumParams() const {
  return sigma + 1;
}

ARDLinearNoise::ARDLinearNoise(const Matrix& data) : 
  Covariance(data.getRows()), ARDLinear(data), sigma(v+3) {
}

void 
ARDLinearNoise::eval(const Vector& param) {
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++)
      (*this)[i][j] = 
	ard(train[i], train[j], param) + linear(train[i], train[j], param);
    (*this)[i][i] += param[sigma];
  }
}

double 
ARDLinearNoise::pred(const Vector& x, const Vector& param, Vector& k) const {
  for(int i = 0; i < dime; i++)
    k[i] = ard(x, train[i], param) + linear(x, train[i], param);
  return ard(x, x, param) + linear(x, x, param) + param[sigma];
}

ARDLinearNoise::~ARDLinearNoise() {
}

////////// ARD + Linear + ARCH //////////

inline
int 
ARDLinearARCH::getNumParams() const {
  return beta + 1;
}

ARDLinearARCH::ARDLinearARCH(const Matrix& data, double s0) :
  Covariance(data.getRows()), ARDLinear(data), sigma0(s0), omega(v+3), beta(v+4) {
}

void 
ARDLinearARCH::eval(const Vector& param) {
  double s0 = sigma0;
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++)
      (*this)[i][j] = 
	ard(train[i], train[j], param) + linear(train[i], train[j], param);
    s0 = param[omega] + param[beta] * s0;
    (*this)[i][i] += s0;
  }
}

double 
ARDLinearARCH::pred(const Vector& x, const Vector& param, Vector& k) const {
  for(int i = 0; i < dime; i++)
    k[i] = ard(x, train[i], param) + linear(x, train[i], param);

  static double s0 = sigma0;
  s0 = param[omega] + param[beta] * s0;
  double b = ard(x, x, param) + linear(x, x, param) + s0;

  return b;
}

ARDLinearARCH::~ARDLinearARCH() {
}

////////// ARD + Linear + Quadratic GARCH //////////

inline
int 
ARDLinearQuadGARCH::getNumParams() const {
  return beta + 1;
}

ARDLinearQuadGARCH::ARDLinearQuadGARCH(const Matrix& data,
				       const Vector& targets, double s0) :
  Covariance(data.getRows()), ARDLinear(data), targ(targets),
  sigma0(s0), omega(v+3), alpha(v+4), beta(v+5) {
}

void 
ARDLinearQuadGARCH::eval(const Vector& param) {
  double s0 = sigma0;
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++)
      (*this)[i][j] = 
	ard(train[i], train[j], param) + linear(train[i], train[j], param);
    s0 = (param[omega] + param[alpha] * Math::sqr(targ[i]) + param[beta]  * s0);
    (*this)[i][i] += s0;
  }
}

double 
ARDLinearQuadGARCH::pred(const Vector& x, const Vector& param, Vector& k) const {
  for(int i = 0; i < dime; i++)
    k[i] = ard(x, train[i], param) + linear(x, train[i], param);

  static int n = 0;
  static double s0 = sigma0;
  s0 = (param[omega] + param[alpha] * Math::sqr(targ[n++]) + param[beta] * s0);
  double b = ard(x, x, param) + linear(x, x, param) + s0;

  return b;
}

ARDLinearQuadGARCH::~ARDLinearQuadGARCH() {
}

////////// ARD + Linear + Quadratic AGARCH //////////

inline
int 
ARDLinearQuadAGARCH::getNumParams() const {
  return beta + 1;
}

ARDLinearQuadAGARCH::ARDLinearQuadAGARCH(const Matrix& data,
					 const Vector& targets, double s0) :
  Covariance(data.getRows()), ARDLinear(data), targ(targets),
  sigma0(s0), omega(v+3), alpha(v+4), rho(v+5), beta(v+6) {
}

void 
ARDLinearQuadAGARCH::eval(const Vector& param) {
  double s0 = sigma0;
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++)
      (*this)[i][j] = 
	ard(train[i], train[j], param) + linear(train[i], train[j], param);
    s0 = (param[omega] + param[alpha] * Math::sqr(targ[i] - param[rho]) +
	  param[beta]  * s0);
    (*this)[i][i] += s0;
  }
}

double 
ARDLinearQuadAGARCH::pred(const Vector& x, const Vector& param, Vector& k) const {
  for(int i = 0; i < dime; i++)
    k[i] = ard(x, train[i], param) + linear(x, train[i], param);

  static int n = 0;
  static double s0 = sigma0;
  s0 = (param[omega] + param[alpha] * Math::sqr(targ[n++] - param[rho]) +
	param[beta]  * s0);
  double b = ard(x, x, param) + linear(x, x, param) + s0;

  return b;
}

ARDLinearQuadAGARCH::~ARDLinearQuadAGARCH() {
}
