#include <cmath>
#include <cassert>
#include "Math.hh"
#include "Random.hh"

inline
void
Random::setSeed(long seed) {
  delete rand1;
  delete rand2;
  init(seed);
}

Random::Random(long seed) : table_size(113), randout(0L) {
  table = new long[table_size];
  init(seed);
}

long
Random::rand() {
  int i = (int)((double)table_size * (double)(randout & 0x7FFFFFFFL) /
		((double)0x7FFFFFFFL + 1.0));
  randout = table[i];
  table[i] = (rand1->rand() << 16) | rand2->rand();
  return randout;
}

double 
Random::uniform() {
  double d = 0x7FFFFFFFL + 1.0;
  double r1 = rand() & 0x7FFFFFFFL;
  double r2 = rand() & 0x7FFFFFFFL;
  return (r1 + r2 / d) / d;
}

double 
Random::uniopen() {
  double u;
  do {
    u = uniform();
  } while(u <= 0.0 || u >= 1.0);
  return u;
}

double
Random::normal() {
  static Boolean extra = FALSE;
  static double d;
  if(extra) {
    extra = FALSE;
    return d;
  } else {
    double u1, u2;
    u1 = sqrt(-2.0 * log(uniopen()));
    u2 = 2.0 * Math::pi() * uniform();
    d = u1 * sin(u2);
    extra = TRUE;
    return u1 * cos(u2);
  }
}

double
Random::gamma(int v) {
#ifdef DEBUG
  assert(v >= 1);
#endif
  if(v == 1) {
    double x = normal();
    return 0.5 * x * x;
  } else if(v == 2) {
    return -log(uniopen());
  } else {
    double x, y, z, vm, root;
    vm = 0.5 * v - 1.0;
    root = sqrt(v - 1.0);
    for(;;) {
      y = tan(Math::pi() * uniform());
      x = root * y + vm;
      if(x <= 0.0) continue;
      z = (1.0 + y * y) * exp(vm * log(x / vm) - root * y);
      if(uniform() <= z) return x;
    }
  }
}

double
Random::beta(int v1, int v2) {
  double x1 = gamma(v1);
  double x2 = gamma(v2);
  return x1 / (x1 + x2);
}

void 
Random::randSphere(Vector& v) {
  double length = 0.0;
  for(int i = 0; i < v.getSize(); i++) {
    v[i] = normal();
    length += v[i] * v[i];
  }
  v /= sqrt(length);
}

void 
Random::cauchy(Vector& v) {
  double x;
  if(v.getSize() == 1) {
    x = Math::pi() * uniform() - 0.5 * Math::pi();
    v[0] = tan(0.99999999 * x);
  } else {
    randSphere(v);
    do {
      x = beta(v.getSize(), 1);
    } while(x >= 1.0);
    x = sqrt(x / (1.0 - x));
    v *= x;
  }
}

Random::~Random() {
  delete [] table;
  delete rand1;
  delete rand2;
}

void
Random::init(long seed) {
  long NUM[10][3] = {
    {214326L, 3613L, 45289L}, {714025L, 1366L, 150889L},
    {134456L, 8121L, 28411L}, {259200L, 7141L, 54773L},
    {244944L, 1597L, 51749L}, {139968L, 3877L, 29573L},
    {312500L, 741L,  66037L}, {145800L, 3661L, 30809L}, 
    {175000L, 2661L, 36979L}, {233280L, 1861L, 49297L}
  };
  int a = (int)Math::abs(seed % 5);
  int b = a + 5;
  rand1 = new Random16(97, NUM[a][0], NUM[a][1], NUM[a][2], (seed >> 16) & 65535L);
  rand2 = new Random16(103,NUM[b][0], NUM[b][1], NUM[b][2], seed & 65535L);
  for(int i = 0; i < table_size; i++)
    table[i] = (rand1->rand() << 16) | rand2->rand();
  randout = (rand1->rand() << 16) | rand2->rand();
}

Random::Random16::Random16(int length, long mm, long aa, long cc, long ss) :
  table_size(length), m(mm), a(aa), c(cc), seed(ss), randout(0L) {
    table = new long[table_size];
    for(int i = 0; i < table_size; i++) {
      seed = (a * seed + c) % m;
      table[i] = seed;
    }
    seed = (a * seed + c) % m;
    randout = seed;
}

long 
Random::Random16::rand() {
  int i = (int)((double)table_size * (double)randout / (double)m);
  randout = table[i];
  seed = (a * seed + c) % m;
  table[i] = seed;
  return randout;
}

Random::Random16::~Random16() {
  delete [] table;
}
