#include <cmath>
#include <cassert>
#include "Math.hh"
#include "SymmetricMatrix.hh"

inline
int
SymmetricMatrix::getDime() const {
  return dime;
}

inline
const
Vector&
SymmetricMatrix::operator[](int i) const {
#ifdef DEBUG
  assert(i >= 0 && i < dime);
#endif
  return vecPtr[i];
}

inline
Vector&
SymmetricMatrix::operator[](int i) {
#ifdef DEBUG
  assert(i >= 0 && i < dime);
#endif
  return vecPtr[i];
}

inline
double
SymmetricMatrix::operator()(int i, int j) const {
#ifdef DEBUG
  assert(i >= 0 && i < dime);
  assert(j >= 0 && j < dime);
#endif
  return ( j <= i ? vecPtr[i][j] : vecPtr[j][i] ); 
}  

inline
double&
SymmetricMatrix::operator()(int i, int j) {
#ifdef DEBUG
  assert(i >= 0 && i < dime);
  assert(j >= 0 && j < dime);
#endif
  return ( j <= i ? vecPtr[i][j] : vecPtr[j][i] ); 
}  

SymmetricMatrix::SymmetricMatrix(int dimensions) :
  dime(dimensions), vecPtr(new Vector[dime]) {
#ifdef DEBUG
    assert(dime >= 0);
#endif
    for(int i = 0; i < dime; i++)
      (*this)[i].resize(i + 1);
}

SymmetricMatrix::SymmetricMatrix(const SymmetricMatrix& matrix) :
  dime(matrix.dime), vecPtr(new Vector[dime]) {
    for(int i = 0; i < dime; i++) {
      (*this)[i].resize(i + 1);
      (*this)[i] = matrix[i];
    }
}

SymmetricMatrix&
SymmetricMatrix::operator=(const SymmetricMatrix& matrix) {
#ifdef DEBUG
  assert(dime == matrix.dime);
#endif
  if(this != &matrix)
    for(int i = 0; i < dime; i++)
      (*this)[i] = matrix[i];
  return *this;
}

SymmetricMatrix&
SymmetricMatrix::operator=(double value) {
  for(int i = 0; i < dime; i++)
    (*this)[i] = value;
  return *this;
}

double
SymmetricMatrix::choleskyDecompose() {
  double s, logDet = 0.0;
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++) {
      s = (*this)[i][j];
      for(int k = j - 1; k >= 0; k--) {
	s -= (*this)[i][k] * (*this)[j][k];
      }
      if(i == j) {
	if(s <= 0.0)
	  throw NotPosDefException();
	logDet += log(s);
	(*this)[i][i] = sqrt(s);
      }
      else {
	(*this)[i][j] = s / (*this)[j][j];
      }
    }
  }
  return logDet;
}

double
SymmetricMatrix::choleskyDecomposeInvert() {
  double s, logDet = choleskyDecompose();
  Vector xVec(dime);
  SymmetricMatrix tmpMat(dime);
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j < dime; j++) {
      s = (i == j) ? 1.0 : 0.0;
      for(int k = j - 1; k >= 0; k--)
	s -= (*this)[j][k] * xVec[k];
      xVec[j] = s / (*this)[j][j];
    }
    for(int j = dime - 1; j >= i; j--) {
      s = xVec[j];
      for(int k = j + 1; k < dime; k++)
	s -= (*this)[k][j] * xVec[k];
      tmpMat[j][i] = xVec[j] = s / (*this)[j][j];
    }
  }
  (*this) = tmpMat;
  return logDet;
}

double
SymmetricMatrix::innovations(const Vector& x, Vector& p, Vector& v) {
#ifdef DEBUG
  assert(x.getSize() == dime);
  assert(p.getSize() == dime);
  assert(v.getSize() == dime);
#endif
  p[0] = 0.0;
  v[0] = (*this)[0][0];
  if(v[0] <= 0.0)
    throw NotPosDefException();
  double s, r, logDet = log(v[0]);
  for(int n = 0; n < dime - 1; n++) {
    for(int k = 0; k <= n; k++) {
      s = 0.0;
      for(int j = 0; j < k; j++)
	s += (*this)[k - 1][k - j - 1] * (*this)[n][n - j] * v[j];
      (*this)[n][n - k] = (1 / v[k]) * ((*this)[n + 1][k] - s);
    }
    s = 0.0; r = 0.0;
    for(int j = 0; j <= n; j++) {
      s += (*this)[n][j] * (x[n - j] - p[n - j]);
      r += Math::sqr((*this)[n][n - j]) * v[j];
    }
    p[n+1] = s;
    v[n+1] = (*this)[n + 1][n + 1] - r;
    if(v[n + 1] <= 0.0)
      throw NotPosDefException();
    logDet += log(v[n + 1]);
  }
  for(int  i = 0; i < v.getSize(); i++)
    v[i] = 1.0 / v[i];
  return logDet;
}

double
SymmetricMatrix::traceProduct(const SymmetricMatrix& matrix) const {
#ifdef DEBUG
  assert(dime == matrix.dime);
#endif
  double sum1 = 0.0, sum2 = 0.0;
  for(int i = 0; i < dime; i++) {
    for(int j = 0; j <= i; j++) {
      if(i == j) 
	sum1 += (*this)[i][i] * matrix[i][i];
      else
	sum2 += (*this)[i][j] * matrix[i][j];
    }
  }
  return sum1 + 2.0 * sum2;
}

double 
SymmetricMatrix::vecXMatXvec(const Vector& vector) const {
#ifdef DEBUG
  assert(dime == vector.getSize());
#endif
  Vector vec(dime);
  vecXMat(vector, vec);
  return vec.dotProduct(vector);
}

void 
SymmetricMatrix::vecXMat(const Vector& vector, Vector& result) const {
#ifdef DEBUG
  assert(dime == vector.getSize());
  assert(dime == result.getSize());
#endif
  result = 0.0;
  for(int i = 0; i < dime; i++) {
    for(int j = i; j < dime; j++)
      result[i] += (*this)[j][i] * vector[j];
  }
  for(int i = 1; i < dime; i++) {
    for(int j = 0; j < i; j++)
      result[i] += (*this)[i][j] * vector[j];
  }
}

SymmetricMatrix::~SymmetricMatrix() {
  delete [] vecPtr;
}

ostream&
operator<<(ostream& oStream, const SymmetricMatrix& matrix) {
  for(int i = 0; i < matrix.dime; i++)
    oStream << matrix[i] << endl;
  return oStream;
}

istream&
operator>>(istream& iStream, SymmetricMatrix& matrix) {
  for(int i = 0; i < matrix.dime; i++)
    iStream >> matrix[i];
  return iStream;
}
