#include <cmath>
#include <cassert>
#include "Math.hh"
#include "Vector.hh"

inline
int
Vector::getSize() const {
  return size;
}

inline
double
Vector::operator[](int i) const {
#ifdef DEBUG
  assert(i >= 0 && i < size);
#endif
  return dblPtr[i];
}

inline
double&
Vector::operator[](int i) {
#ifdef DEBUG
  assert(i >= 0 && i < size);
#endif
  return dblPtr[i];
}

Vector::Vector(int length) : 
  size(length), dblPtr(new double[length]) {
#ifdef DEBUG
  assert(size >= 0);
#endif
}

Vector::Vector(int length, double init) :
  size(length), dblPtr(new double[length]) {
#ifdef DEBUG
  assert(size >= 0);
#endif
  (*this) = init;
}

Vector::Vector(const Vector& vector) :
  size(vector.size), dblPtr(new double[size]) {
  for(int i = 0; i < size; i++)
    (*this)[i] = vector[i];
}

Vector&
Vector::operator=(const Vector& vector) {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  if(this != &vector)
    for(int i = 0; i < size; i++)
      (*this)[i] = vector[i];
  return *this;
}

Vector&
Vector::operator=(double value) {
  for(int i = 0; i < size; i++)
    (*this)[i] = value;
  return *this;
}

Vector& 
Vector::operator+=(double d) {
  for(int i = 0; i < size; i++)
    (*this)[i] += d;
  return *this;
}

Vector& 
Vector::operator-=(double d) {
  for(int i = 0; i < size; i++)
    (*this)[i] -= d;
  return *this;
}

Vector& 
Vector::operator*=(double d) {
  for(int i = 0; i < size; i++)
    (*this)[i] *= d;
  return *this;
}

Vector& 
Vector::operator/=(double d) {
  for(int i = 0; i < size; i++)
    (*this)[i] /= d;
  return *this;
}

Vector 
Vector::operator+(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  Vector vec(size);
  for(int i = 0; i < size; i++)
    vec[i] = (*this)[i] + vector[i];
  return vec;
}

Vector
Vector::operator-(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  Vector vec(size);
  for(int i = 0; i < size; i++)
    vec[i] = (*this)[i] - vector[i];
  return vec;
}

Vector
Vector::operator*(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  Vector vec(size);
  for(int i = 0; i < size; i++)
    vec[i] = (*this)[i] * vector[i];
  return vec;
}

Vector
Vector::operator/(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  Vector vec(size);
  for(int i = 0; i < size; i++)
    vec[i] = (*this)[i] / vector[i];
  return vec;
}

double
Vector::dotProduct(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  double sum = 0.0;
  for(int i = 0; i < size; i++)
    sum += (*this)[i] * vector[i];
  return sum;
}

double
Vector::squaredNorm(const Vector& vector) const {
#ifdef DEBUG
  assert(size == vector.size);
#endif
  double sum = 0.0;
  for(int i = 0; i < size; i++)
    sum += Math::sqr((*this)[i] - vector[i]);
  return sum;
}

double
Vector::squaredNorm(const Vector& vector, const Vector& weights) const {
#ifdef DEBUG
  assert(size == vector.size);
  assert(size <= weights.size);
#endif
  double sum = 0.0;
  for(int i = 0; i < size; i++)
    sum += weights[i] * Math::sqr((*this)[i] - vector[i]);
  return sum;
}

Vector
Vector::slice(int a, int length) const {
#ifdef DEBUG
  assert(a >= 0 && a + length  <= size);
#endif
  Vector vec(length);
  for(int i = 0; i < length; i++)
    vec[i] = (*this)[a + i];
  return vec;
}

void
Vector::resize(int length) {
#ifdef DEBUG
  assert(length >= 0);
#endif
  if(size != length) {
    delete [] dblPtr;
    size = length;
    dblPtr = new double[size];
  }
}

Vector::~Vector() {
  delete [] dblPtr;
}

ostream& 
operator<<(ostream& oStream, const Vector& vector) {
  oStream.precision(8);
  for(int i = 0; i < vector.size; i++) {
    oStream.width(16);
    oStream << vector[i];
  }
  return oStream;
}

istream&
operator>>(istream& iStream, Vector& vector) {
  for(int i = 0; i < vector.size; i++)
    iStream >> vector[i];
  return iStream;
}
