#include <fstream.h>
#include <iostream.h>
#include "Math.hh"
#include "Random.hh"
#include "Vector.hh"
#include "Matrix.hh"
#include "Anneal.hh"
#include "Gnuplot2d.hh"
#include "Posterior.hh"
#include "Covariance.hh"

/*prototypes for functions in this file*/
Matrix read_data(char* fileName);
double posterior(const Vector& param);
void constraint(const Vector& point, Vector& newpoint, double temp);
void predict(Covariance& cov, Matrix& test, Vector& t, 
	     Vector& p, Vector& u, Vector& v);

Vector*     TPtr;  // pointer to target vector
Covariance* COVPtr; // pointer to covariance matrix

void main(void) {
  int TRAIN_N    = 100;// size of training window
  int PRED_N     = 110;// TRAIN_N+size of testing window
  int TARGET_COL = 1;  // target column (the range is from  0 to number of columns in file)
  int SKIP       = 0;  // if user wants to skip some first values in data file

  /*4.dat contains NY stock exchange data*/
  Matrix data = read_data("./data/4.dat");

  /*Inputs to be used for prediction (includes the inputs used for training)*/
  Matrix P(PRED_N, 1);
  
  /*X will contain inputs, the number 1 means that inputs will be scalars but more generally
this number represents the dimensionality of the input vector*/  
Matrix X(TRAIN_N, 1);

/*T will contains some targets (the ones used for training)*/  
Vector T(X.getRows());

/*Y will contain all targets (the ones used for training and the ones needed for testing)*/  
  Vector Y(P.getRows());

/*Targets shifted by 1 (default), used for GARCH and AGARCH as inputs */  
  Vector T_shift(P.getRows());

  /*grid step*/
  double h = 1.0 / (double)P.getRows();
  /*artificial inputs are generated through a grid */
  for(int i = 0; i < P.getRows(); i++) P[i][0] = i * h;
  for(int i = 0; i < X.getRows(); i++) X[i][0] = i * h;
  
  /*load T from data, need SKIP+1 because the shifting by minus 1*/
  T = data.col(TARGET_COL).slice(SKIP + 1, X.getRows());

  Y = data.col(TARGET_COL).slice(SKIP + 1, P.getRows());
  /*the shifting has already taken place given the way T was loaded*/
  T_shift = data.col(TARGET_COL).slice(SKIP, P.getRows());


/*Uncommenting the next 2 lines allows user to use shifted targets as inputs*/
  // P.col(1, T_shift);
  // X.col(1, T_shift.slice(0, X.getRows()));

  /*avg. volatility used as starting value for ARCH-like models*/
  double avgvol = sqrt(T.dotProduct(T) / (double)T.getSize());
 
 // Select a covariance (model) 
  //----------------------------------------------------------------------
  // Construction of covariance matrix for a specific model (see file Covariance.hh)
 // ARDLinearNoise       COV(X);
  // ARDLinearARCH        COV(X, avgvol);
  // ARDLinearQuadGARCH   COV(X, T_shift, avgvol);
  ARDLinearQuadAGARCH  COV(X, T_shift, avgvol);

  /* if one these models was chosen for training, uncomment the define
     for the constrains to take effect*/
  //#define ARCH
  //#define GARCH
#define AGARCH

  TPtr   = &T;
  COVPtr = &COV;


 /* Construct annealing object with function to minimize (posterior) and function constraint that proposes
     new points which satisfy ARCH, GARCH, AGARCH constraints */
  Anneal an(posterior, constraint);
  /* create vector of parameters used by annealing, with initial point supplied (10.0)*/
  Vector param(COV.getNumParams(), 1.0);
  /*annealing returns*/
  an.anneal(param);

  /*construct  two vectors to hold predictions*/
  Vector means(P.getRows()), vars(P.getRows());
  predict(COV, P, T, param, means, vars);

  /*need GNUPLOT below, output two eps files in local directory
   (the .eps files output to consoles if the 2 lines below (POSTCRIPT ) are commented out
NEED TO BE ROOT TO RUN GNUPLOT*/
  Gnuplot2d gp;
  gp.setStyle(LINES);
  gp.setTerminal(POSTSCRIPT, "plot1.eps");
  gp.setTitle("Plot of X vs. Targets and X vs. Predicted Means");
  gp.plot(P.col(0), Y, P.col(0), means);
  gp.message("");
  for(int i = 0; i < vars.getSize(); i++) vars[i] = sqrt(vars[i]);
  gp.setTerminal(POSTSCRIPT, "plot2.eps");
  gp.setTitle("Plot of X vs. Predicted Means with errorbars");
  gp.plot(P.col(0), means, vars);
  gp.message("");
}

double
posterior(const Vector& param) {
  IAMinusLogLike like(*COVPtr, *TPtr);
  double mlp;
  try {
    mlp = like.minusLogLike(param);
  } catch(NotPosDefException& e) {
    mlp = 1E+30;
  }
  return mlp;
}

/*constrains below are messy*/
void 
constraint(const Vector& point, Vector& newpoint, double temp) {
  static Random ran(1283L);
  static double stdev = 50.0;

  int size = point.getSize();
#ifdef ARCH
  size -= 1;
#endif
#ifdef GARCH
  size -= 2;
#endif
#ifdef AGARCH
  size -= 4;
#endif
  for(int i = 0; i < size; i++) {
    do {
      if(temp > 0.15) 
	newpoint[i] = point[i] + temp * stdev * ran.normal();
      else /*if temperature is low use adaptive stdev*/
	newpoint[i] = point[i] + temp * point[i] * ran.normal();
    } while(newpoint[i] <= 0.0);
  }
#ifdef ARCH
  int i = point.getSize() - 1;
  do {
    if(temp > 0.15)
      newpoint[i] = ran.uniform();
    else
      newpoint[i] = point[i] + temp * point[i] * ran.normal();
  } while(newpoint[i] <= 0.0 || newpoint[i] >= 1.0);
#endif
#ifdef GARCH
  int i = point.getSize() - 2;
  do {
    if(temp > 0.15) {
      newpoint[i] = ran.uniform();
      newpoint[i + 1] = ran.uniform();
    } else {
      newpoint[i] = point[i] + temp * stdev * ran.normal();
      newpoint[i + 1] = point[i + 1] + temp * stdev * ran.normal();
    }
  } while(newpoint[i] <= 0.0 || newpoint[i + 1] <= 0.0 ||
	  newpoint[i] + newpoint[i + 1] >= 1.0);
#endif
#ifdef AGARCH
  int b = point.getSize() - 1;
  int r = point.getSize() - 2;
  int a = point.getSize() - 3;
  int w = point.getSize() - 4;
  double rho, omega;
  do {
    if(temp > 0.15) {
      newpoint[a] = ran.uniform();
      newpoint[b] = ran.uniform();
    } else {
      newpoint[a] = point[a] + temp * point[a] * ran.normal();
      newpoint[b] = point[b] + temp * point[b] * ran.normal();
    }
    newpoint[r] = point[r] + temp * point[r] * ran.normal();
    newpoint[w] = point[w] + temp * point[w] * ran.normal();
    rho   = newpoint[r]; 
    omega = newpoint[w];
  } while(newpoint[r] <= 0.0 || newpoint[w] <= 0.0 ||
	  newpoint[a] <= 0.0 || newpoint[b] <= 0.0 ||
	  newpoint[a] + newpoint[b] >= 1.0         ||
	  omega * omega + newpoint[a] * rho * rho >= 
	  (omega + newpoint[a] * rho * rho) / (1 - (newpoint[a] + newpoint[b])));
#endif
}

void 
predict(Covariance& cov, Matrix& test, Vector& t, Vector& p, Vector& u, Vector& v) {
  try {
    cov.eval(p);
    cov.choleskyDecomposeInvert();
  } catch(NotPosDefException& e) {
    cout << "Error: covariance not positive definite" << endl;
    exit(0);
  }
  Vector k(cov.getDime()), kXCov(cov.getDime());
  for(int i = 0; i < test.getRows(); i++) {
    double b = cov.pred(test[i], p, k);
    cov.vecXMat(k, kXCov);
    u[i] = kXCov.dotProduct(t);
    v[i] = b - kXCov.dotProduct(k);
  }
}

/* Read a data file into a matrix (file must be in block format) */
Matrix read_data(char* fileName) {
  ifstream inStream(fileName);
  if(inStream == 0) {
    cerr << "Unable to open file: " << fileName << endl;
    exit(0);
  }
  int rows = 0, cols = 0, n = 0;
  double dbl;
  while(inStream.peek() != '\n' && inStream.peek() != EOF) {
    inStream >> dbl;
    cols++;
  }
  inStream.close();
  inStream.open(fileName);
  while(inStream.peek() != EOF) {
    if(inStream.peek() == '\n') {
      rows++;
      if(cols != n) {
	cerr << "In file: " << fileName << endl;  
	cerr << "There are " << n << " column(s) on line " << rows + 1 << endl;
	cerr << "Was expecting " << cols << endl;
	exit(0);
      }
      n = 0;
    }
    inStream >> dbl;
    if(inStream.fail() && !inStream.eof()) {
      cerr << "In file: " << fileName << endl;  
      cerr << "Invalid characters in file at line: " << rows + 1 << endl;
      exit(0);
    }
    n++;
  }
  inStream.close();
  inStream.open(fileName);
  Matrix mat(rows, cols);
  inStream >> mat;
  return mat;
}















