#include <fstream.h>
#include <iostream.h>
#include "Math.hh"
#include "Random.hh"
#include "Vector.hh"
#include "Matrix.hh"
#include "Anneal.hh"
#include "Gnuplot2d.hh"
#include "Posterior.hh"
#include "Covariance.hh"

// Function Prototypes
double posterior(const Vector& param);
void constraint(const Vector& point, Vector& newpoint, double temp);
void genData(Covariance& cov, Matrix& x, Vector& t, Vector& param);
void predict(Covariance& cov, Matrix& test, Vector& t, Vector& p, Vector& u, Vector& v);
//

Vector*     TPtr;   // pointer to target vector
Covariance* COVPtr; // pointer to covariance matrix

// Vector of additional inputs for GARCH and AGARCH models (kicks to volatility)
Vector T_gen(128);

void main(void) {
  Matrix X(128, 1);      // inputs - 128 rows, 1 column (dimensionality of input vector = DIM)
                          //this main assumes DIM is 1 (in the generation of inputs)
  Vector T(X.getRows()); // target vector

  // Select a covariance (model) to generate targets
  //----------------------------------------------------------------------
  // Construction of covariance matrix for a specific model (see file Covariance.hh)
  ARDLinearNoise      COV_gen(X);
  // ARDLinearARCH       COV_gen(X, 5.0); // arguments: input matrix and inital volatility

  // note: GARCH and AGARCH are passed T_gen 
  // ARDLinearQuadGARCH  COV_gen(X, T_gen, 5.0); 
  // ARDLinearQuadAGARCH COV_gen(X, T_gen, 5.0);
 
  // fix parameters (we are simulating data!). The order of parameters (for each Model) is as follows:
  // | w | v | a0 | a1 | sigma | for ARD+Linear+white noise
  //there is only one w because DIM =1
  // 
  // | w | v | a0 | a1 | omega | beta | for ARD+Linear+ARCH noise
  //there is only one w because DIM =1
  // | w | v | a0 | a1 | omega | alpha | beta | for ARD+Linear+GARCH noise
  //there is only one w because DIM =1
  // | w | v | a0 | a1 | omega | alpha | rho | beta | for ARD+Linear+AGARCH noise
  //there is only one w because DIM =1
 
  Vector true_param(COV_gen.getNumParams(), 0.0); //0.0 to initialize true_param vector
  // these first four paraemeters are common to all models
  /* w  */ true_param[0] = 100.0;
  /* v  */ true_param[1] = 10.0;
  /* a0 */ true_param[2] = 20.0;
  /* a1 */ true_param[3] = 20.0;
  
  // noise parameter
  true_param[4] = 1e-4;
  
  //true_param[5] = 0.5;
  //true_param[6] = 0.5;
  //true_param[7] = 0.5;

  
  TPtr   = &T;
  COVPtr = &COV_gen;
  
  /* generate data, with zero mean*/
  genData(COV_gen, X, T, true_param);
  
  /* output value of posterior at chosen parameters */
  double mlp_true = posterior(true_param);
  cout << mlp_true << endl;

  /* Attempt to recover parameters */
  //----------------------------------------------------------------------

  /* estimate average stdev from simulated data, needed for ARCHC-like models*/
  double avgvol = sqrt(T.dotProduct(T) / (double)T.getSize());

  /* choose a model to train with */
  ARDLinearNoise       COV(X) ;
  //ARDLinearARCH        COV(X, avgvol);
  //ARDLinearQuadGARCH   COV(X, T_gen, avgvol);
  //ARDLinearQuadAGARCH  COV(X, T_gen, avgvol);

  /* if one these models was chosen for training, uncomment the define
     for the constrains to take effect*/
  //#define ARCH
  //#define GARCH
  //#define AGARCH

  /* set COVPtr to new covariance model */
  COVPtr = &COV;

  /* Construct annealing object with function to minimize (posterior) and function 
     constraint that proposes
     new points which satisfy ARCH, GARCH, AGARCH constraints */
  Anneal an(posterior, constraint);
  
  /* create vector of parameters used by annealing, with initial point supplied (10.0)*/
  Vector param(COV.getNumParams(), 10.0);
  /*annealing returns*/
  an.anneal(param);

  /*construct  two vectors to hold predictions*/
  Vector means(X.getRows()), vars(X.getRows());
  predict(COV, X, T, param, means, vars);

  /*output learned parmeters and errors*/
  //----------------------------------------------------------------------
  double mlp = posterior(param);
  cout << "Posterior achieved: " << mlp                   << endl;
  cout << "Posterior true:     " << mlp_true              << endl;
  cout << "Relative error (posterior):"                   << endl; 
  cout << Math::abs(mlp - mlp_true) / Math::abs(mlp_true) << endl;

  cout << "Parameters found: " << param                   << endl;
  cout << "Parameters true:  " << true_param              << endl;
  cout << "Relative error (parameters):"                  << endl;
  if(param.getSize() == true_param.getSize())
    for(int i = 0; i < param.getSize(); i++)
      cout << Math::abs(param[i]-true_param[i])/Math::abs(true_param[i]) << endl;

  /*need GNUPLOT below, output two eps files in local directory
   (the .eps files output to consoles if the 2 lines below (POSTCRIPT ) are commented out
NEED TO BE ROOT TO RUN GNUPLOT*/
  Gnuplot2d gp;
  gp.setTerminal(POSTSCRIPT, "plot1.eps");
  gp.setTitle("Plot of X vs. Targets and X vs. Predicted Means");
  gp.plot(X.col(0), T, X.col(0), means);
  gp.message("");
  for(int i = 0; i < vars.getSize(); i++) vars[i] = sqrt(vars[i]);
  gp.setTerminal(POSTSCRIPT, "plot2.eps");
  gp.setTitle("Plot of X vs. Predicted Means with errorbars");
  gp.plot(X.col(0), means, vars);
  gp.message("");
  //----------------------------------------------------------------------
}

double
posterior(const Vector& param) {
  IAMinusLogLike like(*COVPtr, *TPtr);
  double mlp;
  try {
    mlp = like.minusLogLike(param);
  } catch(NotPosDefException& e) {
    mlp = 1E+30;
  }
  return mlp;
}

/*constrains below are messy*/
void 
constraint(const Vector& point, Vector& newpoint, double temp) {
  static Random ran(1283L);
  static double stdev = 50.0;

  int size = point.getSize();
#ifdef ARCH
  size -= 1;
#endif
#ifdef GARCH
  size -= 2;
#endif
#ifdef AGARCH
  size -= 4;
#endif
  for(int i = 0; i < size; i++) {
    do {
      if(temp > 0.15) 
	newpoint[i] = point[i] + temp * stdev * ran.normal();
      else /*if temperature is low use adaptive stdev*/
	newpoint[i] = point[i] + temp * point[i] * ran.normal();
    } while(newpoint[i] <= 0.0);
  }
#ifdef ARCH
  int i = point.getSize() - 1;
  do {
    if(temp > 0.15)
      newpoint[i] = ran.uniform();
    else
      newpoint[i] = point[i] + temp * point[i] * ran.normal();
  } while(newpoint[i] <= 0.0 || newpoint[i] >= 1.0);
#endif
#ifdef GARCH
  int i = point.getSize() - 2;
  do {
    if(temp > 0.15) {
      newpoint[i] = ran.uniform();
      newpoint[i + 1] = ran.uniform();
    } else {
      newpoint[i] = point[i] + temp * stdev * ran.normal();
      newpoint[i + 1] = point[i + 1] + temp * stdev * ran.normal();
    }
  } while(newpoint[i] <= 0.0 || newpoint[i + 1] <= 0.0 ||
	  newpoint[i] + newpoint[i + 1] >= 1.0);
#endif
#ifdef AGARCH
  int b = point.getSize() - 1;
  int r = point.getSize() - 2;
  int a = point.getSize() - 3;
  int w = point.getSize() - 4;
  double rho, omega;
  do {
    if(temp > 0.15) {
      newpoint[a] = ran.uniform();
      newpoint[b] = ran.uniform();
    } else {
      newpoint[a] = point[a] + temp * point[a] * ran.normal();
      newpoint[b] = point[b] + temp * point[b] * ran.normal();
    }
    newpoint[r] = point[r] + temp * point[r] * ran.normal();
    newpoint[w] = point[w] + temp * point[w] * ran.normal();
    rho   = newpoint[r]; 
    omega = newpoint[w];
  } while(newpoint[r] <= 0.0 || newpoint[w] <= 0.0 ||
	  newpoint[a] <= 0.0 || newpoint[b] <= 0.0 ||
	  newpoint[a] + newpoint[b] >= 1.0         ||
	  omega * omega + newpoint[a] * rho * rho >= 
	  (omega + newpoint[a] * rho * rho) / (1 - (newpoint[a] + newpoint[b])));
#endif
}

void 
genData(Covariance& cov, Matrix& x, Vector& t, Vector& param) {
  double h = 1.0 / (double)x.getRows();
  for(int i = 0; i < x.getRows(); i++) x[i][0] = i * h;
  T_gen = x.col(0);

  try {
    cov.eval(param);
    cov.choleskyDecompose();
  } catch(NotPosDefException& e) {
    cout << "Error: covariance not positive definite" << endl;
    exit(0);
  }

  Matrix chol(cov);
  for(int i = 0; i < chol.getRows(); i++)
    for(int j = i+1; j < chol.getCols(); j++)
      chol[i][j] = 0.0;

  Random ran(797L);
  Vector e(chol.getRows());
  for(int i = 0; i < e.getSize(); i++)
    e[i] = ran.normal();

  t = chol * e;
  
  Gnuplot2d gp;
  gp.plot(x.col(0), t);
  gp.message("Plot of x vs. Targets");

  // data file output if desired, we are not testing with this main
  ofstream train("train.0");
  ofstream test("test.0");
  Matrix xt(x.getRows(), x.getCols()+1);
  xt.col(0, x.col(0));
  xt.col(1, t);
  train << xt;
  test << x;
}

void 
predict(Covariance& cov, Matrix& test, Vector& t, Vector& p, Vector& u, Vector& v) {
  try {
    cov.eval(p);
    cov.choleskyDecomposeInvert();
  } catch(NotPosDefException& e) {
    cout << "Error: covariance not positive definite" << endl;
    exit(0);
  }
  
  /*loop for predictions*/
  Vector k(cov.getDime()), kXCov(cov.getDime());
  for(int i = 0; i < test.getRows(); i++) {
    double b = cov.pred(test[i], p, k);
    cov.vecXMat(k, kXCov);
    u[i] = kXCov.dotProduct(t);
    v[i] = b - kXCov.dotProduct(k);
  }
}






