#include <fstream.h>
#include <iostream.h>
#include "Random.hh"
#include "Vector.hh"
#include "Matrix.hh"
#include "Anneal.hh"
#include "Gnuplot2d.hh"
#include "Posterior.hh"
#include "Covariance.hh"

double posterior(const Vector& param);
void constraint(const Vector& point, Vector& newpoint, double temp);
Matrix read_data(char* fileName);

Vector*     TPtr;
Covariance* COVPtr;

void main(void) {
  char* DATA_FILE_NAME = "./data/1.dat";
  int   TARGET_COL     = 1;
  int   SKIP_N         = 0;
  int   WINDOW_N       = 256;
  int   HISTORY_N      = 1;
  int   SLIDE_N        = 16;

  Matrix data = read_data(DATA_FILE_NAME);

  Matrix X(WINDOW_N, HISTORY_N);
  Vector T(WINDOW_N); 

  T = data.col(TARGET_COL).slice(SKIP_N + HISTORY_N, WINDOW_N);

  // cout << "Train T:\n" << T << endl;

  for(int i = 0; i < HISTORY_N; i++)
    X.col(i, data.col(TARGET_COL).slice(SKIP_N + i, WINDOW_N));

  // cout << "Train X:\n" << X;

  // data file output
  ofstream train("train.0");
  Matrix xt(X.getRows(), X.getCols()+1);
  xt.col(0, X.col(0));
  xt.col(1, T);
  train << xt;
  exit(0);
  //

  double avgvol = sqrt(T.dotProduct(T) / (double)T.getSize());
  // ARDLinearNoise          COV(X);
  // ARDLinearARCH        COV(X, avgvol);
  // ARDLinearQuadGARCH   COV(X, avgvol);
  ARDLinearQuadAGARCH  COV(X, avgvol);

  TPtr   = &T;
  COVPtr = &COV;

  Anneal an(posterior, constraint);
  Vector param(COV.getNumParams(), 1.0);
  an.anneal(param);

  Vector means(SLIDE_N), vars(SLIDE_N);
  for(int i = 0; i < SLIDE_N; i++) {

    T = data.col(TARGET_COL).slice(SKIP_N + HISTORY_N + i, WINDOW_N);

    // cout << "T:\n" << T << endl;

    for(int j = 0; j < HISTORY_N; j++)
      X.col(j, data.col(TARGET_COL).slice(SKIP_N + i + j, WINDOW_N));

    // cout << "X:\n" << X;

    COV.eval(param);
    COV.choleskyDecomposeInvert();

    Vector test(HISTORY_N);
    test = T.slice(WINDOW_N - HISTORY_N, HISTORY_N);

    // cout << "test:\n" << test << endl;

    Vector k(COV.getDime()), kXCov(COV.getDime());
    double b = COV.pred(test, param, k);
    COV.vecXMat(k, kXCov);
    means[i] = kXCov.dotProduct(T);
    vars[i] = b - kXCov.dotProduct(k);
  }

  Gnuplot2d gp;
  Vector x(SLIDE_N), y(SLIDE_N);
  for(int i = 0; i < SLIDE_N; i++) x[i] = i;
  gp.setStyle(LINES);

  y = data.col(TARGET_COL).slice(SKIP_N + WINDOW_N + HISTORY_N, SLIDE_N);

  // cout << "p:\n" << y << endl;

  gp.setTerminal(POSTSCRIPT, "plot1.eps");
  gp.setTitle("Plot of X vs. Y and X vs. Predicted Means");
  gp.plot(x, y, x, means);
  gp.message("");
  for(int i = 0; i < SLIDE_N; i++) vars[i] = sqrt(vars[i]);
  gp.setTerminal(POSTSCRIPT, "plot2.eps");
  gp.setTitle("Plot of X vs. Predicted Means with Error Bars");
  gp.plot(x, means, vars);
  gp.message("");

  int n = 0;
  for(int i = 0; i < SLIDE_N - 1; i++) {
    if(y[i] - y[i+1] <= 0.0 && means[i] - means[i+1] <= 0.0)      n++;
    else if(y[i] - y[i+1] >= 0.0 && means[i] - means[i+1] >= 0.0) n++;
  }
  
  cout << "Percent: " << (double)n / (double)(SLIDE_N - 1) << endl;
}

double
posterior(const Vector& param) {
  IAMinusLogLike like(*COVPtr, *TPtr);
  double mlp;
  try {
    mlp = like.minusLogLike(param);
  } catch(NotPosDefException& e) {
    mlp = 1E+30;
  }
  return mlp;
}

void 
constraint(const Vector& point, Vector& newpoint, double temp) {
  static Random ran(1283L);
  static double stdev = 5.0;
  for(int i = 0; i < point.getSize(); i++) {
    do {
      if(temp > 0.15)
	newpoint[i] = point[i] + temp * stdev * ran.normal();
      else
	newpoint[i] = point[i] + temp * point[i] * ran.normal();
    } while(newpoint[i] <= 0.0);
  }
}

Matrix read_data(char* fileName) {
  ifstream inStream(fileName);
  if(inStream == 0) {
    cerr << "Unable to open file: " << fileName << endl;
    exit(0);
  }
  int rows = 0, cols = 0, n = 0;
  double dbl;
  while(inStream.peek() != '\n' && inStream.peek() != EOF) {
    inStream >> dbl;
    cols++;
  }
  inStream.close();
  inStream.open(fileName);
  while(inStream.peek() != EOF) {
    if(inStream.peek() == '\n') {
      rows++;
      if(cols != n) {
	cerr << "In file: " << fileName << endl;  
	cerr << "There are " << n << " column(s) on line " << rows + 1 << endl;
	cerr << "Was expecting " << cols << endl;
	exit(0);
      }
      n = 0;
    }
    inStream >> dbl;
    if(inStream.fail() && !inStream.eof()) {
      cerr << "In file: " << fileName << endl;  
      cerr << "Invalid characters in file at line: " << rows + 1 << endl;
      exit(0);
    }
    n++;
  }
  inStream.close();
  inStream.open(fileName);
  Matrix mat(rows, cols);
  inStream >> mat;
  return mat;
}
