#include <fstream.h>
#include <iostream.h>
#include "Math.hh"
#include "Random.hh"
#include "Vector.hh"
#include "Matrix.hh"
#include "Anneal.hh"
#include "Gnuplot2d.hh"
#include "Posterior.hh"
#include "Covariance.hh"

double 
posterior(const Vector& param);
void 
constraint(const Vector& point, Vector& newpoint, double temp);
void
genData(Covariance& cov, Matrix& x, Vector& t, Vector& param);
void
predict(Covariance& cov, Matrix& test, Vector& t, Vector& p, Vector& u, Vector& v);

Vector*     TPtr;
Covariance* COVPtr;

void main(void) {
  Matrix X(256, 1);
  Vector T(X.getRows()); 

  ARDLinearNoise COV(X);
  // ARDLinearARCH  COV(X, 5.0);

  Vector true_param(COV.getNumParams());
  /* w0    */ true_param[0] = 10.0;
  /* v     */ true_param[1] = 10.0;
  /* a0    */ true_param[2] = 20.0;
  /* a1    */ true_param[3] = 20.0;
  /* sigma */ true_param[4] = 0.001;
  /* beta  */ // true_param[5] = 0.95;

  genData(COV, X, T, true_param);

  TPtr   = &T;
  COVPtr = &COV;

  Anneal an(posterior, constraint);
  Vector param(COV.getNumParams(), 100.0);
  an.anneal(param);

  Vector means(X.getRows()), vars(X.getRows());
  predict(COV, X, T, param, means, vars);

  Gnuplot2d gp;
  gp.setTerminal(POSTSCRIPT, "plot1.eps");
  gp.setTitle("Plot of X vs. Targets and X vs. Predicted Means");
  gp.plot(X.col(0), T, X.col(0), means);
  gp.message("");
  for(int i = 0; i < vars.getSize(); i++) vars[i] = sqrt(vars[i]);
  gp.setTerminal(POSTSCRIPT, "plot2.eps");
  gp.setTitle("Plot of X vs. Predicted Means with errorbars");
  gp.plot(X.col(0), means, vars);
  gp.message("");

  double mlp_a = posterior(param);
  double mlp_b = posterior(true_param);
  cout << "Posterior achieved: " << mlp_a             << endl;
  cout << "Posterior true:     " << mlp_b             << endl;
  cout << "Relative error (posterior):"               << endl; 
  cout << Math::abs(mlp_a - mlp_b) / Math::abs(mlp_b) << endl;
  cout << "Parameters found: " << param               << endl;
  cout << "Parameters true:  " << true_param          << endl;
  cout << "Relative error (parameters):"              << endl; 
  for(int i = 0; i < param.getSize(); i++)
    cout << Math::abs(param[i] - true_param[i]) / Math::abs(true_param[i]) << endl; 
}

double
posterior(const Vector& param) {
  IAMinusLogLike like(*COVPtr, *TPtr);
  double mlp;
  try {
    mlp = like.minusLogLike(param);
  } catch(NotPosDefException& e) {
    mlp = 1E+30;
  }
  return mlp;
}

void 
constraint(const Vector& point, Vector& newpoint, double temp) {
  static Random ran(1283L);
  static double stdev = 50.0;
  for(int i = 0; i < point.getSize(); i++) {
    do {
      if(temp > 0.15)
	newpoint[i] = point[i] + temp * stdev * ran.normal();
      else
	newpoint[i] = point[i] + temp * point[i] * ran.normal();
    } while(newpoint[i] <= 0.0);
  }
}

void 
genData(Covariance& cov, Matrix& x, Vector& t, Vector& param) {
  double h = 1.0 / (double)x.getRows();
  for(int i = 0; i < x.getRows(); i++) x[i][0] = i * h;

  try {
    cov.eval(param);
    cov.choleskyDecompose();
  } catch(NotPosDefException& e) {
    cout << "Error: covariance not positive definite" << endl;
    exit(0);
  }

  Matrix chol(cov);
  for(int i = 0; i < chol.getRows(); i++)
    for(int j = i+1; j < chol.getCols(); j++)
      chol[i][j] = 0.0;

  Random ran(797L);
  Vector e(chol.getRows());
  for(int i = 0; i < e.getSize(); i++)
    e[i] = ran.normal();

  t = chol * e;
  
  Gnuplot2d gp;
  gp.plot(x.col(0), t);
  gp.message("Plot of x vs. Targets");

  // data file output
  ofstream train("train.0");
  ofstream test("test.0");
  Matrix xt(x.getRows(), x.getCols()+1);
  xt.col(0, x.col(0));
  xt.col(1, t);
  train << xt;
  test << x;
}

void 
predict(Covariance& cov, Matrix& test, Vector& t, Vector& p, Vector& u, Vector& v) {
  try {
    cov.eval(p);
    cov.choleskyDecomposeInvert();
  } catch(NotPosDefException& e) {
    cout << "Error: covariance not positive definite" << endl;
    exit(0);
  }
  Vector k(cov.getDime()), kXCov(cov.getDime());
  for(int i = 0; i < test.getRows(); i++) {
    double b = cov.pred(test[i], p, k);
    cov.vecXMat(k, kXCov);
    u[i] = kXCov.dotProduct(t);
    v[i] = b - kXCov.dotProduct(k);
  }
}
